import logging
import threading
from contextvars import ContextVar, Token
from typing import Optional

from pyxxl.schema import RunData


class GlobalVars:
    _DATA: ContextVar = ContextVar("_DATA")
    _LOGGER: ContextVar = ContextVar("_LOGGER")
    _EVENT: ContextVar = ContextVar("_EVENT")

    @classmethod
    def set_xxl_run_data(cls, data: RunData) -> None:
        cls._DATA.set(data)

    @classmethod
    def try_get_run_data(cls) -> Optional[RunData]:
        return cls._DATA.get(None)

    @property
    def xxl_run_data(self) -> RunData:
        return self._DATA.get()

    @classmethod
    def set_task_logger(cls, logger: logging.Logger) -> Token:
        return cls._LOGGER.set(logger)

    @property
    def logger(self) -> logging.Logger:  # pragma: no cover
        return self._LOGGER.get()

    @classmethod
    def set_cancel_event(cls, event: threading.Event) -> None:
        cls._EVENT.set(event)

    @property
    def cancel_event(self) -> threading.Event:  # pragma: no cover
        return self._EVENT.get()


g = GlobalVars()



import asyncio
import time
from typing import Iterator

import pytest

from pyxxl.enum import executorBlockStrategy
from pyxxl.error import JobDuplicateError, JobNotFoundError, JobParamsError
from pyxxl.executor import Executor, JobHandler
from pyxxl.schema import RunData

job_handler = JobHandler()


@job_handler.register
async def pytest_executor_async():
    await asyncio.sleep(2)
    return "成功30"


@job_handler.register
def pytest_executor_sync():
    time.sleep(2)
    return "成功30"


@job_handler.register
async def pytest_executor_error():
    assert 1 == 2


HANDLER_NAMES = [
    "pytest_executor_async",
    "pytest_executor_sync",
]


@pytest.mark.asyncio
async def test_runner_not_found(executor: Executor, job_id: int, log_id: int):
    executor.reset_handler(job_handler)
    with pytest.raises(JobNotFoundError):
        await executor.run_job(
            RunData.from_dict(
                dict(
                    logId=log_id,
                    jobId=job_id,
                    executorHandler="not_found",
                    executorBlockStrategy=executorBlockStrategy.DISCARD_LATER.value,
                    errorTest="errorTest",
                )
            )
        )
    await executor.graceful_close()


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_callback(executor: Executor, handler_name: str):
    executor.reset_handler(job_handler)
    executor.xxl_client.clear_result()
    data = RunData.from_dict(
        dict(
            logId=1,
            jobId=11,
            executorHandler=handler_name,
            executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
        )
    )
    await executor.run_job(data)
    data = RunData.from_dict(
        dict(
            logId=2,
            jobId=12,
            executorHandler="pytest_executor_error",
            executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
        )
    )
    await executor.run_job(data)
    await executor.graceful_close()
    assert executor.xxl_client.callback_result.get(1) == 200
    assert executor.xxl_client.callback_result.get(2) == 500
    assert executor.xxl_client.callback_result.get(3) is None


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_cancel(executor: Executor, handler_name: str):
    executor.reset_handler(job_handler)
    cancel_job_id, ok_job_id = 1100, 1200
    cancel_log_id, ok_log_id = 1100, 1200
    base_data = dict(
        executorHandler=handler_name,
        executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
    )
    await executor.run_job(RunData(logId=cancel_log_id, jobId=cancel_job_id, **base_data))
    await executor.run_job(RunData(logId=ok_log_id, jobId=ok_job_id, **base_data))

    await executor.cancel_job(cancel_job_id, include_queue=False)
    await executor.graceful_close(10)
    assert executor.xxl_client.callback_result.get(cancel_job_id) == 500
    assert executor.xxl_client.callback_result.get(ok_job_id) == 200


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_cancel_include_queue(
    executor: Executor, handler_name: str, job_id: int, log_id_iter: Iterator[int]
):
    executor.reset_handler(job_handler)
    base_data = dict(
        jobId=job_id,
        executorHandler=handler_name,
        executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
    )
    cancel_log_id, queue_log_id = next(log_id_iter), next(log_id_iter)
    await executor.run_job(RunData(logId=cancel_log_id, **base_data))
    await executor.run_job(RunData(logId=queue_log_id, **base_data))
    await executor.cancel_job(job_id, include_queue=True)
    await executor.graceful_close(10)
    assert executor.xxl_client.callback_result.get(cancel_log_id) == 500
    assert executor.xxl_client.callback_result.get(queue_log_id) is None


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_SERIAL_EXECUTION(executor: Executor, job_id: int, handler_name: str, log_id_iter: Iterator[int]):
    executor.xxl_client.clear_result()
    executor.reset_handler(job_handler)
    run_data = dict(
        jobId=job_id,
        executorHandler=handler_name,
        executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
    )
    queue_size = 3
    log_ids = [next(log_id_iter) for _ in range(queue_size)]
    for log_id in log_ids:
        await executor.run_job(RunData(logId=log_id, **run_data))

    assert executor.queue.get(job_id).qsize() == queue_size - 1
    await executor.graceful_close(10)
    assert executor.queue.get(job_id).qsize() == 0
    assert executor.xxl_client.callback_result.get(log_id) == 200

    # max_queue_length
    for _ in range(executor.config.task_queue_length + 1):
        await executor.run_job(RunData(logId=next(log_id_iter), **run_data))

    with pytest.raises(JobDuplicateError, match="discard"):
        await executor.run_job(RunData(logId=next(log_id_iter), **run_data))

    await executor.shutdown()


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_DISCARD_LATER(executor: Executor, job_id: int, handler_name: str, log_id_iter: Iterator[int]):
    executor.reset_handler(job_handler)
    executor.xxl_client.clear_result()
    run_data = dict(
        jobId=job_id,
        executorHandler=handler_name,
        executorBlockStrategy=executorBlockStrategy.DISCARD_LATER.value,
    )
    ok_log_id, duplicate_log_id = next(log_id_iter), next(log_id_iter)
    await executor.run_job(RunData(logId=ok_log_id, **run_data))
    with pytest.raises(JobDuplicateError):
        await executor.run_job(RunData(logId=duplicate_log_id, **run_data))
    await executor.graceful_close(10)
    assert executor.xxl_client.callback_result.get(ok_log_id) == 200
    assert executor.xxl_client.callback_result.get(duplicate_log_id) is None


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_COVER_EARLY(executor: Executor, job_id: int, handler_name: str, log_id_iter: Iterator[int]):
    executor.reset_handler(job_handler)
    executor.xxl_client.clear_result()
    run_data = dict(
        jobId=job_id,
        executorHandler=handler_name,
        executorBlockStrategy=executorBlockStrategy.COVER_EARLY.value,
    )
    ok_log_id, error_log_id = next(log_id_iter), next(log_id_iter)
    await executor.run_job(RunData(logId=error_log_id, **run_data))
    await executor.run_job(RunData(logId=ok_log_id, **run_data))
    await executor.graceful_close(10)
    assert executor.xxl_client.callback_result.get(ok_log_id) == 200
    assert executor.xxl_client.callback_result.get(error_log_id) == 500


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_OTHER(executor: Executor, job_id: int, handler_name: str, log_id_iter: Iterator[int]):
    executor.reset_handler(job_handler)
    with pytest.raises(JobParamsError, match="unknown executorBlockStrategy"):
        for _ in range(2):
            await executor.run_job(
                RunData(
                    logId=next(log_id_iter),
                    jobId=job_id,
                    executorHandler=handler_name,
                    executorBlockStrategy="OTHER",
                )
            )
    executor.xxl_client.clear_result()


@pytest.mark.asyncio
async def test_sync_timeout(executor: Executor, job_id: int, log_id: int):
    from pyxxl.ctx import g

    sync_handler = JobHandler()

    @sync_handler.register(name="pytest_executor_sync")
    def pytest_executor_sync():
        while not g.cancel_event.is_set():
            time.sleep(1)

    executor.reset_handler(sync_handler)
    await executor.run_job(
        RunData(
            logId=log_id,
            jobId=job_id,
            executorHandler="pytest_executor_sync",
            executorBlockStrategy="OTHER",
            executorTimeout=2,
        )
    )
    await executor.graceful_close(10)
    assert executor.xxl_client.callback_result.get(log_id) == 500



from typing import Any


class JobDuplicateError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class JobNotFoundError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class JobParamsError(Exception):
    def __init__(self, message: str, **kwargs: Any) -> None:
        self.message = message + ", ".join(["[%s=%s]" % (k, v) for k, v in kwargs.items()])
        super().__init__(message)


class JobRegisterError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class XXLClientError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)



import asyncio
import time

import pytest

from pyxxl.ctx import g
from pyxxl.executor import HandlerInfo


@pytest.mark.asyncio
async def test_async_timeout():
    r = []

    async def _handler():
        while True:
            await asyncio.sleep(0.5)
            r.append(1)

    handler = HandlerInfo(handler=_handler)
    with pytest.raises(asyncio.TimeoutError):
        await handler.start(2)

    await asyncio.sleep(1)
    now_r_num = len(r)
    await asyncio.sleep(2)
    assert len(r) == now_r_num


@pytest.mark.asyncio
async def test_sync_timeout_ok(event_loop):
    r = []

    def _handler():
        while len(r) < 10 and not g.cancel_event.is_set():
            time.sleep(0.5)
            r.append(1)

    handler = HandlerInfo(handler=_handler)
    with pytest.raises(asyncio.TimeoutError):
        await handler.start(2)

    await asyncio.sleep(1)
    now_r_num = len(r)
    await asyncio.sleep(2)
    assert len(r) == now_r_num


@pytest.mark.asyncio
async def test_sync_timeout_error(event_loop):
    r = []

    def _handler():
        while len(r) < 10:  # 防止测试线程卡死
            time.sleep(0.5)
            r.append(1)

    handler = HandlerInfo(handler=_handler)
    with pytest.raises(asyncio.TimeoutError):
        await handler.start(2)

    await asyncio.sleep(1)
    now_r_num = len(r)
    await asyncio.sleep(2)
    assert len(r) > now_r_num



from flask import Flask

app = Flask(__name__)


@app.route("/")
def hello_world():
    return "<p>Hello, World!</p>"


if __name__ == "__main__":
    # 如果是多进程部署的flask无法这样使用，请参考gunicorn的example
    # 此案例仅用于本地开发和调试用
    from multiprocessing import freeze_support

    from executor_app import app as pyxxl_app

    freeze_support()

    app.pyxxl_app = pyxxl_app
    pyxxl_app.run_with_daemon()
    app.run()



import time
from pathlib import Path
from typing import TYPE_CHECKING, Callable

import aiofiles
import pytest

from pyxxl.logger import DiskLog, LogBase, RedisLog
from pyxxl.tests.utils import INSTALL_REDIS, REDIS_TEST_URI
from pyxxl.types import LogRequest, LogResponse
from pyxxl.utils import try_import

if TYPE_CHECKING:
    import redis
else:
    redis = try_import("redis")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "get_log",
    [
        lambda: DiskLog("", log_tail_lines=20),
        pytest.param(
            lambda: RedisLog("pyxxl-test", REDIS_TEST_URI, log_tail_lines=20),
            marks=pytest.mark.skipif(not INSTALL_REDIS, reason="no redis package."),
        ),
    ],
    ids=["disk", "redis"],
)
@pytest.mark.parametrize(
    "req,resp",
    [
        (
            LogRequest(logDateTim=0, logId=0, fromLineNum=1),
            LogResponse(
                fromLineNum=1,
                toLineNum=20,
                logContent="".join(str(i) + "\n" for i in range(1, 21)),
                isEnd=False,
            ),
        ),
        (
            LogRequest(logDateTim=0, logId=0, fromLineNum=21),
            LogResponse(
                fromLineNum=21,
                toLineNum=40,
                logContent="".join(str(i) + "\n" for i in range(21, 41)),
                isEnd=False,
            ),
        ),
        (
            LogRequest(logDateTim=0, logId=0, fromLineNum=81),
            LogResponse(
                fromLineNum=81,
                toLineNum=81,
                logContent="",
                isEnd=True,
            ),
        ),
    ],
)
async def test_read_file(get_log: Callable[..., LogBase], req, resp):
    log = get_log()
    data = [str(i).encode() + b"\n" for i in range(1, 80 + 1)]
    async with log.mock_write(*data) as key:
        assert await log.get_logs(req, key=key) == resp


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "get_log",
    [
        lambda: DiskLog("", log_tail_lines=20),
        pytest.param(
            lambda: RedisLog("pyxxl-test", REDIS_TEST_URI, log_tail_lines=20),
            marks=pytest.mark.skipif(not INSTALL_REDIS, reason="no redis package."),
        ),
        pytest.param(
            lambda: RedisLog("pyxxl-test", redis.ConnectionPool.from_url(REDIS_TEST_URI), log_tail_lines=20),
            marks=pytest.mark.skipif(not INSTALL_REDIS, reason="no redis package."),
        ),
    ],
    ids=["disk", "redis", "redis-pool"],
)
async def test_disk_logger(get_log: Callable[..., LogBase]):
    log = get_log()
    log_id = int(time.time())
    async with log.mock_logger(log_id) as mock_log:
        logger = mock_log.get_logger(log_id)
        try:
            raise ValueError("test error")
        except ValueError as e:
            logger.error(e, exc_info=True)
        logger.warning("test warning.")
        logger.handlers.clear()

        read_data = await mock_log.read_task_logs(log_id)
        for b in ["test error", "test warning", "ERROR", "WARNING"]:
            assert b in read_data
    # test file not found
    assert (
        "No such logid logs"
        in (
            await log.get_logs(
                LogRequest(logDateTim=0, logId=0, fromLineNum=81),
                key="xxxxxxxxxxxx.log",
            )
        )["logContent"]
    )


@pytest.mark.asyncio
async def test_disk_expired():
    log_id = int(time.time())
    async with aiofiles.tempfile.TemporaryDirectory() as d:
        file_log = DiskLog(log_path=d, expired_days=0)
        logger = file_log.get_logger(log_id, stdout=False)
        logger.error("test error.")
        logger.warning("test warning.")
        logger.handlers.clear()

        log_file = Path(file_log.key(log_id))
        assert log_file.exists()
        await file_log.expired_once()
        assert log_file.exists() is False



import asyncio

from fastapi import FastAPI

from pyxxl import JobHandler

app = FastAPI()
xxl_handler = JobHandler()


@app.get("/")
async def root():
    return {"message": "Hello World"}


@xxl_handler.register(name="demoJobHandler")
async def test_task():
    await asyncio.sleep(10)
    return "成功10"



import asyncio
import time

from pyxxl import ExecutorConfig, PyxxlRunner
from pyxxl.ctx import g

# 如果xxl-admin可以直连executor的ip，可以不填写executor_listen_host
config = ExecutorConfig(
    xxl_admin_baseurl="http://localhost:8080/xxl-job-admin/api/",
    executor_app_name="xxl-job-executor-sample",
    executor_listen_host="127.0.0.1",  # xxl-admin监听时绑定的host,默认为第一个网卡地址
    debug=True,
)

app = PyxxlRunner(config)


@app.register(name="demoJobHandler")
async def test_task():
    # you can get task params with "g"
    g.logger.info("get executor params: %s" % g.xxl_run_data.executorParams)
    for i in range(10):
        g.logger.warning("test logger %s" % i)
    await asyncio.sleep(5)
    return "成功..."


@app.register(name="xxxxx")
async def test_task3():
    await asyncio.sleep(3)
    return "成功3"


@app.register(name="sync_func")
def test_task4():
    # 如果要在xxl-admin上看到执行日志，打印日志的时候务必用g.logger来打印，默认只打印info及以上的日志
    n = 1
    g.logger.info("Job %s get executor params: %s" % (g.xxl_run_data.jobId, g.xxl_run_data.executorParams))
    # 如果同步任务里面有循环，为了支持cancel操作，必须每次都判断g.cancel_event.
    while n <= 10 and not g.cancel_event.is_set():
        # 如果不需要从xxl-admin中查看日志，可以用自己的logger
        g.logger.info(
            "log to {} logger test_task4.{},params:{}".format(
                g.xxl_run_data.jobId,
                n,
                g.xxl_run_data.executorParams,
            )
        )
        time.sleep(2)
        n += 1
    return "成功3"


if __name__ == "__main__":
    app.run_executor()



import logging
from contextlib import contextmanager
from typing import Generator

from pyxxl.ctx import g

from .common import LogBase
from .disk import DiskLog
from .redis import RedisLog


@contextmanager
def new_logger(factory: LogBase, log_id: int) -> Generator[logging.Logger, None, None]:
    logger = factory.get_logger(log_id)
    token = g.set_task_logger(logger)
    yield logger
    factory.after_running(logger)
    g._LOGGER.reset(token)



import asyncio
import logging
from typing import Any, Dict, List, Optional, Union

import aiohttp
from yarl import URL

from pyxxl.error import XXLClientError
from pyxxl.log import xxl_client_logger

JsonType = Union[None, int, str, bool, List[Any], Dict[Any, Any]]


class Response:
    def __init__(self, code: int, msg: Optional[str] = None, **kwargs: Any) -> None:
        self.code = code
        self.msg = msg

    @property
    def ok(self) -> bool:
        return self.code == 200


class XXL:
    def __init__(
        self,
        admin_url: str,
        token: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        retry_times: int = 1,
        retry_interval: int = 5,
        session: Optional[aiohttp.ClientSession] = None,
        logger: Optional[logging.Logger] = None,
        **kwargs: Any,
    ) -> None:
        self.loop = loop or asyncio.get_event_loop()
        kwargs["loop"] = self.loop

        _admin_url: URL = URL(admin_url)
        if not (_admin_url.scheme.startswith("http") and _admin_url.path.endswith("/")):
            raise ValueError("admin_url must like http://localhost:8080/xxl-job-admin/api/")

        # https://docs.aiohttp.org/en/stable/client_reference.html#baseconnector
        self.url_path = _admin_url.path
        if not session:  # for pytest
            self.conn = aiohttp.TCPConnector(**kwargs)
            session = aiohttp.ClientSession(
                base_url=_admin_url.origin(),
                connector=self.conn,
                trust_env=True,
            )

        self.session = session

        self.retry_times = retry_times
        self.retry_interval = retry_interval
        self.headers = {"XXL-JOB-ACCESS-TOKEN": token} if token else {}
        self.logger = logger or xxl_client_logger

    async def registry(self, key: str, value: str) -> bool:
        payload = dict(registryGroup="EXECUTOR", registryKey=key, registryValue=value)
        try:
            await self._post("registry", payload, retry_times=5)
            return True
        except XXLClientError as e:
            self.logger.error("Registry executor failed. %s", e.message)
        return False

    async def registryRemove(self, key: str, value: str) -> None:
        payload = dict(registryGroup="EXECUTOR", registryKey=key, registryValue=value)
        await self._post("registryRemove", payload, retry_times=3)
        self.logger.info("RegistryRemove successful. %s" % payload)

    async def callback(self, log_id: int, timestamp: int, code: int = 200, msg: Optional[str] = None) -> None:
        # executeResult兼容xxl-job2.2版本
        payload = [
            {
                "logId": log_id,
                "logDateTim": timestamp,
                "handleCode": code,
                "handleMsg": msg,
                "executeResult": {"code": code, "msg": msg},
            }
        ]
        await self._post("callback", payload)
        self.logger.debug("Callback successful. %s" % payload)

    async def _post(self, path: str, payload: JsonType, retry_times: Optional[int] = None) -> Response:
        self.logger.debug("post to xxl-job path={} payload={}".format(path, payload))
        times = 0
        retry_times = retry_times or self.retry_times
        while times < retry_times:
            try:
                async with self.session.post(self.url_path + path, json=payload, headers=self.headers) as response:
                    if response.status == 200:
                        r = Response(**(await response.json()))
                        if not r.ok:
                            raise XXLClientError(r.msg or "")
                        return r
                    raise XXLClientError(await response.text())
            except aiohttp.ClientConnectionError as e:
                times += 1
                self.logger.warning(
                    "Connection error {} times, retry after {}. {}".format(times, self.retry_interval, str(e))
                )
                await asyncio.sleep(self.retry_interval)

        raise XXLClientError("Connection error after retry times {}".format(times))

    async def close(self) -> None:
        await self.session.close()
        self.logger.info("http session is closed.")



import pytest

from pyxxl.error import JobRegisterError
from pyxxl.executor import Executor


# pylint: disable=function-redefined
@pytest.mark.asyncio
async def test_hander_error(executor: Executor):
    executor.reset_handler()
    with pytest.raises(JobRegisterError):

        @executor.handler.register
        def test_dup_error(): ...

        @executor.handler.register
        def test_dup_error():  # noqa: F811
            ...


# pylint: disable=function-redefined
@pytest.mark.asyncio
async def test_hander(executor: Executor):
    executor.reset_handler()

    @executor.handler.register
    def test_hander1(): ...

    @executor.handler.register(replace=True)
    async def test_hander1():  # noqa: F811
        ...

    @executor.handler.register(name="test_hander1_dup")
    def test_hander1():  # noqa: F811
        ...



import asyncio
import logging
from abc import ABC, abstractmethod
from typing import Any, AsyncContextManager, Optional

from pyxxl.ctx import g
from pyxxl.types import LogRequest, LogResponse

MAX_LOG_TAIL_LINES = 1000
TASK_FORMAT = (
    "%(asctime)s.%(msecs)03d [%(threadName)s] [%(logId)s] "
    "%(levelname)s %(pathname)s(%(funcName)s:%(lineno)d) - %(message)s"
)
TASKDATE_FORMAT = "%Y-%m-%d %H:%M:%S"
TASK_FORMATTER = logging.Formatter(TASK_FORMAT, datefmt=TASKDATE_FORMAT)


class LogBase(ABC):
    executor_logger: logging.Logger

    @abstractmethod
    def get_logger(self, log_id: int, *, stdout: bool = True, level: int = logging.INFO) -> logging.Logger: ...

    @abstractmethod
    async def get_logs(self, request: LogRequest, *, key: Optional[str] = None) -> LogResponse: ...

    @abstractmethod
    async def read_task_logs(self, log_id: int, *, key: Optional[str] = None) -> str:
        """一次性读取某个log id的所有日志,主要用于单测"""
        ...

    @abstractmethod
    def mock_write(self, *lines: Any) -> AsyncContextManager[str]: ...

    @abstractmethod
    def mock_logger(self, log_id: int) -> AsyncContextManager["LogBase"]: ...

    async def expired_once(self) -> None:  # noqa: B027
        """执行一次批量过期操作,如果是redis啥的自带过期就无需实现此方法"""
        pass

    async def expired_loop(self, seconds: int = 3600) -> None:
        """
        Args:
            seconds (int, optional): one loop seconds. Defaults to 3600.
        """
        self.executor_logger.debug("start expired_loop...")
        try:
            while True:
                await self.expired_once()
                await asyncio.sleep(seconds)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            self.executor_logger.exception(e)
        finally:
            self.executor_logger.info("expired_loop exit...")

    def after_running(self, logger: logging.Logger) -> None:
        return None


class PyxxlFileHandler(logging.FileHandler):
    def emit(self, record: Any) -> None:
        xxl_kwargs = g.try_get_run_data()
        record.logId = xxl_kwargs.logId if xxl_kwargs else "NotInTask"
        return super().emit(record)


class PyxxlStreamHandler(logging.StreamHandler):
    def emit(self, record: Any) -> None:
        xxl_kwargs = g.try_get_run_data()
        record.logId = xxl_kwargs.logId if xxl_kwargs else "NotInTask"
        return super().emit(record)



import atexit
from multiprocessing.util import _exit_function

from pyxxl import ExecutorConfig, PyxxlRunner

bind = ["0.0.0.0:8000"]
backlog = 512
workers = 1
timeout = 300
graceful_timeout = 2
limit_request_field_size = 8192


def when_ready(server):
    # pylint: disable=import-outside-toplevel,unused-import,no-name-in-module
    from app import xxl_handler

    atexit.unregister(_exit_function)

    config = ExecutorConfig(
        xxl_admin_baseurl="http://localhost:8080/xxl-job-admin/api/",
        executor_app_name="xxl-job-executor-sample",
        executor_listen_host="172.17.0.1",
        debug=True,
    )

    pyxxl_app = PyxxlRunner(config, handler=xxl_handler)
    server.pyxxl_app = pyxxl_app
    pyxxl_app.run_with_daemon()



import os
from urllib.parse import urlparse

import pytest

from pyxxl.setting import ExecutorConfig
from pyxxl.utils import get_network_ip

TEST_ADMIN_URL = "http://localhost:8080/xxl-job-admin/api/"


def test_config():
    setting = ExecutorConfig(
        xxl_admin_baseurl=TEST_ADMIN_URL,
        executor_app_name="test",
        dotenv_try=False,
    )
    assert urlparse(setting.executor_url).hostname == get_network_ip()
    assert setting.executor_app_name == "test"

    # like nginx proxy
    setting = ExecutorConfig(
        xxl_admin_baseurl=TEST_ADMIN_URL,
        executor_app_name="test",
        executor_url="http://nginx_domain",
        dotenv_try=False,
    )
    assert setting.executor_listen_port == 9999
    assert setting.executor_listen_host == get_network_ip()
    assert setting.executor_app_name == "test"

    # from env
    os.environ["executor_app_name"] = "fromenv"
    os.environ["XXL_ADMIN_BASEURL"] = TEST_ADMIN_URL
    os.environ["GRACEFUL_TIMEOUT"] = "500"
    os.environ["GRACEFUL_CLOSE"] = "False"

    setting = ExecutorConfig(
        xxl_admin_baseurl="",
        executor_app_name="",
        dotenv_try=True,
    )
    assert setting.executor_app_name == "fromenv"
    assert setting.xxl_admin_baseurl == TEST_ADMIN_URL
    assert setting.graceful_timeout == 500
    assert setting.graceful_close is False
    os.environ.clear()


@pytest.mark.parametrize(
    "msg,error,kwargs",
    [
        ("admin_url", ValueError, dict(xxl_admin_baseurl="dddd", executor_app_name="test")),
        ("executor_app_name", ValueError, dict(xxl_admin_baseurl=TEST_ADMIN_URL, executor_app_name="")),
        (
            "log_local_dir",
            ValueError,
            dict(
                xxl_admin_baseurl=TEST_ADMIN_URL,
                executor_app_name="test",
                log_target="disk",
                log_local_dir="",
            ),
        ),
        (
            "log_redis_uri",
            ValueError,
            dict(
                xxl_admin_baseurl=TEST_ADMIN_URL,
                executor_app_name="test",
                log_target="redis",
                log_redis_uri="",
            ),
        ),
    ],
)
def test_error(msg, error, kwargs):
    with pytest.raises(error, match=msg):
        ExecutorConfig(**kwargs, dotenv_try=False)


def test_host_and_port():
    setting = ExecutorConfig(
        xxl_admin_baseurl=TEST_ADMIN_URL,
        executor_app_name="test",
        executor_listen_port=8800,
        dotenv_try=False,
    )
    assert setting.executor_listen_port == 8800
    assert setting.executor_listen_host == get_network_ip()
    assert setting.executor_url == f"http://{get_network_ip()}:8800"

    setting = ExecutorConfig(
        xxl_admin_baseurl=TEST_ADMIN_URL,
        executor_app_name="test",
        executor_listen_port=8800,
        executor_listen_host="pyxxl.com",
        dotenv_try=False,
    )
    assert setting.executor_listen_port == 8800
    assert setting.executor_listen_host == "pyxxl.com"
    assert setting.executor_url == "http://pyxxl.com:8800"

    setting = ExecutorConfig(
        xxl_admin_baseurl=TEST_ADMIN_URL,
        executor_app_name="test",
        executor_listen_port=8800,
        executor_listen_host="pyxxl.com",
        executor_url="http://nginx-domain:8080",
        dotenv_try=False,
    )
    assert setting.executor_listen_port == 8800
    assert setting.executor_listen_host == "pyxxl.com"
    assert setting.executor_url == "http://nginx-domain:8080"



from __future__ import annotations

import asyncio
import logging
import threading
import time
import warnings
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, MutableSet, Optional

from pyxxl import error
from pyxxl.ctx import g
from pyxxl.enum import executorBlockStrategy
from pyxxl.log import executor_logger
from pyxxl.logger import DiskLog, LogBase, new_logger
from pyxxl.schema import RunData
from pyxxl.setting import ExecutorConfig
from pyxxl.types import DecoratedCallable
from pyxxl.xxl_client import XXL

# https://docs.python.org/3.10/library/asyncio-task.html#asyncio.create_task
_BACKGROUND_TASKS: MutableSet[asyncio.Task] = set()


def _spawn_task(task: asyncio.Task) -> None:
    _BACKGROUND_TASKS.add(task)
    task.add_done_callback(_BACKGROUND_TASKS.discard)


@dataclass
class HandlerInfo:
    handler: Callable
    is_async: bool = False

    def __str__(self) -> str:
        return "<HandlerInfo {}>".format(self.handler.__name__)

    def __post_init__(self) -> None:
        self.is_async = asyncio.iscoroutinefunction(self.handler)

    async def start(self, timeout: int) -> Any:
        if self.is_async:
            return await asyncio.wait_for(self.handler(), timeout=timeout)
        # https://stackoverflow.com/questions/71416383/python-asyncio-cancelling-a-to-thread-task-wont-stop-the-thread
        # 由于线程无法直接取消，这里发送一个event，供开发者自己接收信号来判断是否需要取消
        event = threading.Event()
        g.set_cancel_event(event)
        try:
            return await asyncio.wait_for(asyncio.to_thread(self.handler), timeout=timeout)
        except (asyncio.exceptions.TimeoutError, asyncio.CancelledError) as e:
            event.set()
            # logger.debug("Get error for sync task {}".format(self))
            raise e


class XXLTask:
    def __init__(self, task: asyncio.Task, data: RunData):
        self.task = task
        self.data = data

    def __str__(self) -> str:
        return "<XXLTask task={} data={}>".format(self.task, self.data)

    @property
    def cancel(self) -> Any:
        return self.task.cancel


class JobHandler:
    def __init__(self, logger: Optional[logging.Logger] = None) -> None:
        self._handlers: Dict[str, HandlerInfo] = {}
        self.logger = logger or executor_logger

    def register(
        self, *args: Any, name: Optional[str] = None, replace: bool = False
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        """将函数注册到可执行的job中,如果其他地方要调用该方法,replace修改为True"""

        def func_wrapper(func: DecoratedCallable) -> DecoratedCallable:
            handler_name = name or func.__name__
            if handler_name in self._handlers and replace is False:
                raise error.JobRegisterError("handler %s already registered." % handler_name)
            handler = HandlerInfo(handler=func)
            if not handler.is_async:
                warnings.warn(
                    "Using the sync method will unknown blocking exception, consider using async method.",
                    SyntaxWarning,
                    stacklevel=2,
                )
            self._handlers[handler_name] = handler
            self.logger.debug("register job %s,is async: %s" % (handler_name, asyncio.iscoroutinefunction(func)))

            return func

        if len(args) == 1:
            return func_wrapper(args[0])

        return func_wrapper

    def get(self, name: str) -> Optional[HandlerInfo]:
        return self._handlers.get(name, None)

    def handlers_info(self) -> List[str]:
        return ["<%s is_async:%s>" % (k, v.is_async) for k, v in self._handlers.items()]


class Executor:
    def __init__(
        self,
        xxl_client: XXL,
        config: ExecutorConfig,
        *,
        handler: Optional[JobHandler] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        logger_factory: Optional[LogBase] = None,
        successed_callback: Optional[Callable] = None,
        failed_callback: Optional[Callable] = None,
    ) -> None:
        """执行器，真正的调度任务和策略都在这里

        Args:
            xxl_client (XXL): xxl客户端
            config (ExecutorConfig): 配置参数
            handler (Optional[JobHandler], optional): Defaults to None.
            loop (Optional[asyncio.AbstractEventLoop], optional): Defaults to None.
        """

        self.xxl_client = xxl_client
        self.config = config

        self.handler: JobHandler = handler or JobHandler()
        self.loop = loop or asyncio.get_event_loop()
        self.tasks: Dict[int, XXLTask] = {}
        self.queue: Dict[int, asyncio.Queue[RunData]] = defaultdict(
            lambda: asyncio.Queue(maxsize=self.config.task_queue_length)
        )
        # todo: lock for jobId
        self.lock = asyncio.Lock()
        self.thread_pool = ThreadPoolExecutor(
            max_workers=self.config.max_workers,
            thread_name_prefix="pyxxl_pool",
        )
        self.logger_factory = logger_factory or DiskLog(self.config.log_local_dir)
        self.successed_callback = successed_callback or (lambda: 1)
        self.failed_callback = failed_callback or (lambda x: 1)
        self.loop.set_default_executor(self.thread_pool)

    @property
    def executor_logger(self) -> logging.Logger:
        return self.config.executor_logger

    def _create_task(self, data: RunData) -> XXLTask:
        """创建一个任务"""
        task = self.loop.create_task(self._run(data), name=f"{data.jobId}_{data.logId}")
        return XXLTask(task, data)

    async def run_job(self, data: RunData) -> str:
        handler_obj = self.handler.get(data.executorHandler)
        if not handler_obj:
            self.executor_logger.warning("handler %s not found." % data.executorHandler)
            raise error.JobNotFoundError("handler %s not found." % data.executorHandler)

        # 一个执行器同时只能执行一个jobId相同的任务
        async with self.lock:
            current_task = self.tasks.get(data.jobId)
            if not current_task and self.get_queue(data.jobId).empty():
                self.tasks[data.jobId] = self._create_task(data)
                return "Running"

            self.executor_logger.warning("jobId {} is running. current_task={}".format(data.jobId, current_task))
            # pylint: disable=no-else-raise
            if data.executorBlockStrategy == executorBlockStrategy.DISCARD_LATER.value:
                raise error.JobDuplicateError(
                    "The same job [%s] is already executing and this has been discarded." % data
                )
            elif data.executorBlockStrategy == executorBlockStrategy.COVER_EARLY.value:
                msg = "Job {} BlockStrategy is COVER_EARLY, logId {} replaced.".format(data.jobId, data.logId)
                self.executor_logger.warning(msg)
                await self.get_queue(data.jobId).put(data)
                _spawn_task(self.loop.create_task(self.cancel_job(data.jobId, include_queue=False)))
                return msg
            elif data.executorBlockStrategy == executorBlockStrategy.SERIAL_EXECUTION.value:
                queue = self.get_queue(data.jobId)
                if queue.full():
                    msg = "Job {job_id} is  SERIAL, queue length more than {maxsize}.Job {job}  discard!".format(
                        job_id=data.jobId, job=data, maxsize=queue.maxsize
                    )
                    self.executor_logger.error(msg)
                    raise error.JobDuplicateError(msg)
                else:
                    msg = "job {job_id} is in queen, logId {log_id} ranked {ranked}th [max={maxsize}]...".format(
                        job_id=data.jobId, log_id=data.logId, ranked=queue.qsize() + 1, maxsize=queue.maxsize
                    )
                    self.executor_logger.info(msg)
                    await queue.put(data)
                    return msg
            else:
                raise error.JobParamsError(
                    "unknown executorBlockStrategy [%s]." % data.executorBlockStrategy,
                    executorBlockStrategy=data.executorBlockStrategy,
                )

    async def cancel_job(self, job_id: int, include_queue: bool = True) -> None:
        await asyncio.sleep(0.01)  # delay for pytest
        self.executor_logger.warning("start kill job: job_id={}".format(job_id))
        async with self.lock:
            if include_queue:
                queue = self.get_queue(job_id)
                while not queue.empty():
                    data = queue.get_nowait()
                    self.executor_logger.warning("Discard jobId {} from queue,data: {}".format(job_id, data))

            task = self.tasks.get(job_id, None)
            if task:
                # https://docs.python.org/3/library/asyncio-task.html#asyncio.Task.cancel
                task.cancel()
                try:
                    await task.task
                except asyncio.CancelledError:
                    self.executor_logger.warning("Job %s cancelled." % job_id)

    async def is_running(self, job_id: int) -> bool:
        return job_id in self.tasks

    async def _run(self, data: RunData) -> None:
        handler = self.handler.get(data.executorHandler)
        assert handler
        g.set_xxl_run_data(data)
        with new_logger(self.logger_factory, data.logId) as task_logger:
            start_time = int(time.time() * 1000)
            try:
                task_logger.info("Start job jobId=%s logId=%s [%s]" % (data.jobId, data.logId, data))
                timeout = data.executorTimeout or self.config.task_timeout
                result = await handler.start(timeout)
                task_logger.info("Job finished jobId=%s logId=%s" % (data.jobId, data.logId))
                await self.xxl_client.callback(data.logId, start_time, code=200, msg=result)
                self.successed_callback()
            except asyncio.CancelledError as e:
                task_logger.info(e, exc_info=True)
                await self.xxl_client.callback(data.logId, start_time, code=500, msg="CancelledError")
                self.failed_callback("cancelled")
            except asyncio.exceptions.TimeoutError as e:
                # 同步任务run_in_executor超时会抛出TimeoutError异常
                # !!! 但是注意线程里面的任务仍然在运行，可能会占满所有的线程池
                task_logger.warning(e, exc_info=True)
                await self.xxl_client.callback(data.logId, start_time, code=500, msg="TimeoutError")
                self.failed_callback("timeout")
            except Exception as err:  # pylint: disable=broad-except
                task_logger.exception(err, exc_info=True)
                await self.xxl_client.callback(data.logId, start_time, code=500, msg=str(err))
                self.failed_callback("exception")
            finally:
                if self.lock.locked():
                    await self._finish(data.jobId)
                else:
                    async with self.lock:
                        await self._finish(data.jobId)

    async def _finish(self, job_id: int) -> None:
        # 所有移除tasks的操作全部在这里执行
        finish_task = self.tasks.pop(job_id, None)
        self.executor_logger.info("Finish task {}".format(finish_task))
        queue = self.get_queue(job_id)
        if not queue.empty():
            data = queue.get_nowait()
            self.executor_logger.info(
                "Get data from queue jobId={}, after queueSize={}, data={}".format(job_id, queue.qsize(), data)
            )
            self.tasks[job_id] = self._create_task(data)
            queue.task_done()

    async def shutdown(self) -> None:
        await asyncio.sleep(0.01)  # sleep for pytest

        async with self.lock:
            self.queue.clear()
            for _, task in self.tasks.items():
                task.task.cancel()

    async def graceful_close(self, timeout: int = 60) -> None:
        """优雅关闭"""
        await asyncio.sleep(0.01)  # sleep for pytest

        async def _graceful_close() -> None:
            while len(self.tasks) > 0 or any(i.qsize() > 0 for i in self.queue.values()):
                await asyncio.wait([i.task for i in self.tasks.values()])
                await asyncio.sleep(0.05)

        await asyncio.wait_for(_graceful_close(), timeout=timeout)

    def reset_handler(self, handler: Optional[JobHandler] = None) -> None:
        self.handler = handler or JobHandler()

    def get_queue(self, job_id: int) -> asyncio.Queue[RunData]:
        return self.queue[job_id]



import uuid
from pathlib import Path

from pyxxl.utils import setup_logging, try_import


def test_import():
    assert try_import("pandas") is None
    assert try_import("aiohttp")


def test_logging():
    log_path = "logs/test.log"
    Path(log_path).unlink(missing_ok=True)
    test_log_record = uuid.uuid4().hex
    logger = setup_logging(log_path, "pyxxl-pytest")
    logger.info(test_log_record)
    with open(log_path) as f:
        data = f.readline()
        assert test_log_record in data
        assert "INFO" in data



import os
from typing import Any, Dict, Optional

from pyxxl.main import PyxxlRunner
from pyxxl.utils import try_import
from pyxxl.xxl_client import XXL, JsonType, Response


class MokeXXL(XXL):
    callback_result: Dict[int, Any] = {}

    async def callback(self, log_id: int, timestamp: int, code: int = 200, msg: Optional[str] = None) -> None:
        self.callback_result[log_id] = code

    async def _post(self, path: str, payload: JsonType, retry_times: Optional[int] = None) -> Response:
        return Response(code=200)

    def clear_result(self) -> None:
        self.callback_result = {}


class MokePyxxlRunner(PyxxlRunner):
    def _get_xxl_clint(self) -> MokeXXL:
        return MokeXXL(self.config.xxl_admin_baseurl, token=self.config.access_token)


REDIS_TEST_URI = os.environ.get("REDIS_TEST_URI", "redis://localhost")
INSTALL_REDIS = bool(try_import("redis"))



from typing import Any, Callable, TypedDict, TypeVar

Handler0 = Callable[[Any, Any], Any]
Handler1 = Callable[..., Any]

DecoratedCallable = TypeVar("DecoratedCallable", Handler0, Handler1)


class LogResponse(TypedDict):
    fromLineNum: int
    toLineNum: int
    logContent: str
    isEnd: bool


class LogRequest(TypedDict):
    logDateTim: int
    logId: int
    fromLineNum: int  # min 1



import asyncio
import logging
import os
import time
from contextlib import asynccontextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, AsyncGenerator, List, Optional

import aiofiles

from pyxxl.log import executor_logger
from pyxxl.types import LogRequest, LogResponse

from .common import TASK_FORMATTER, LogBase, PyxxlFileHandler, PyxxlStreamHandler

if TYPE_CHECKING:
    from logging import Handler


LOG_NAME_PREFIX = "pyxxl-{log_id}.log"
LOG_NAME_REGEX = "pyxxl-*.log"
MAX_LOG_TAIL_LINES = 1000


class DiskLog(LogBase):
    def __init__(
        self,
        log_path: str,
        log_tail_lines: int = 0,
        expired_days: float = 14,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        self.log_path = Path(log_path)
        self.executor_logger = logger or executor_logger
        if not self.log_path.exists():
            self.log_path.mkdir()  # pragma: no cover
            self.executor_logger.info("create logdir %s" % self.log_path)  # pragma: no cover
        self.log_tail_lines = log_tail_lines or MAX_LOG_TAIL_LINES
        self.expired_seconds = round(3600 * 24 * expired_days)

    def key(self, log_id: int) -> str:
        return self.log_path.joinpath(LOG_NAME_PREFIX.format(log_id=log_id)).absolute().as_posix()

    def get_logger(self, log_id: int, *, stdout: bool = True, level: int = logging.INFO) -> logging.Logger:
        logger = logging.getLogger("pyxxl.task_log.disk.task-{%s}" % log_id)
        logger.propagate = False
        logger.setLevel(level)
        handlers: list[Handler] = [PyxxlStreamHandler()] if stdout else []
        handlers.append(PyxxlFileHandler(self.key(log_id), delay=True))
        for h in handlers:
            h.setFormatter(TASK_FORMATTER)
            h.setLevel(level)
            logger.addHandler(h)
        return logger

    async def get_logs(self, request: LogRequest, *, key: Optional[str] = None) -> LogResponse:
        # todo: 优化获取中间行的逻辑，缓存之前每行日志的大小然后直接seek
        logs = ""
        to_line_num = request["fromLineNum"]  # start with 1
        is_end = False
        key = key or self.key(request["logId"])
        try:
            async with aiofiles.open(key, mode="r") as f:
                for i in range(1, request["fromLineNum"] + self.log_tail_lines):
                    log = await f.readline()
                    if log == "":
                        is_end = True
                        break
                    elif i >= request["fromLineNum"]:
                        to_line_num = i
                        logs += log
        except FileNotFoundError as e:
            self.executor_logger.warning(str(e), exc_info=True)
            logs = "No such logid logs."

        return LogResponse(
            fromLineNum=request["fromLineNum"],
            toLineNum=to_line_num,
            logContent=logs,
            isEnd=is_end,
        )

    async def read_task_logs(self, log_id: int, *, key: Optional[str] = None) -> str:
        key = key or self.key(log_id)
        async with aiofiles.open(key, mode="r") as f:
            return await f.read()

    async def expired_once(self, batch: int = 1000) -> None:
        now = time.time()
        expire_timestamp = now - self.expired_seconds

        del_list = await asyncio.to_thread(self._scan_expired_files, expire_timestamp)
        self.executor_logger.info("Search expired logs, found %s", len(del_list))

        for i, p in enumerate(del_list, 1):
            if i % batch == 0:
                self.executor_logger.debug("Delete expired logs step: %s", i)
                await asyncio.sleep(0.01)  # release CPU for other tasks
            p.unlink(missing_ok=True)

        if del_list:
            self.executor_logger.info("Delete expired logs successfully, count: %s", len(del_list))

    def _scan_expired_files(self, expire_timestamp: float) -> List[Path]:
        if not self.log_path.exists():
            return []

        del_list: List[Path] = []
        self.executor_logger.debug("Start scan expired files")
        for sub_path in (i for i in self.log_path.glob(LOG_NAME_REGEX) if i.is_file()):
            try:
                ctime = os.path.getctime(sub_path.absolute())
                if ctime < expire_timestamp:
                    del_list.append(sub_path)
            except (OSError, FileNotFoundError) as e:
                # 文件可能在扫描过程中被删除
                self.executor_logger.debug(f"Skip file {sub_path}: {e}")
                continue
        return del_list

    @asynccontextmanager
    async def mock_write(self, *lines: Any) -> AsyncGenerator[str, None]:
        async with aiofiles.tempfile.NamedTemporaryFile() as f:
            await f.writelines(lines)
            await f.flush()
            await f.seek(0)
            yield str(f.name)

    @asynccontextmanager
    async def mock_logger(self, _log_id: int) -> AsyncGenerator[LogBase, None]:
        async with aiofiles.tempfile.TemporaryDirectory() as d:
            handler = DiskLog(log_path=d)
            yield handler

    def after_running(self, logger: logging.Logger) -> None:
        # !!! StreamHandler remove会有问题，需要判断是否是FileHandler
        file_handlers = [h for h in logger.handlers if isinstance(h, logging.FileHandler)]
        for fh in file_handlers:
            logger.debug("close file log object: {}.".format(fh))
            fh.close()
            logger.removeHandler(fh)
